@@ -2,13 +2,18 @@ from numpy.testing import *
 from environment import Settings
 import os
 
+import numpy
+
+import algopy
 from algopy.tracer.tracer import *
 from algopy.utpm import UTPM
-
 from algopy import dot, eigh, qr, trace, solve, inv
 
+try:
+    import mpmath
+except ImportError:
+    mpmath = None
 
-import numpy
 
 class Test_Function_on_numpy_types(TestCase):
 
@@ -180,7 +185,7 @@ class Test_Function_on_UTPM(TestCase):
 
 
     def test_prod(self):
-        tmp = numpy.array([4.*10**-1,10**-20,3*10**10])
+        tmp = numpy.array([4e-1, 1e-20, 3e10])
 
         cg = CGraph()
         x = Function(tmp)
@@ -524,6 +529,30 @@ class Test_CGgraph_on_UTPM(TestCase):
             assert_almost_equal(numpy.trace(numpy.dot(Ab.T,Ad)), numpy.trace(numpy.dot(Qb.T,Qd) + numpy.dot(Rb.T,Rd)))
 
 
+    def test_pullback5(self):
+        cg = CGraph()
+        x = UTP([3.])
+        y = UTP([7.])
+        fx = Function(x)
+        fy = Function(y)
+        fv1 = fx * fy
+        fv2 = (fv1 * fx + fy)*fv1
+        cg.independentFunctionList = [fx,fy]
+        cg.dependentFunctionList = [fv2]
+
+        v2bar = UTP([1.])
+        cg.pullback([v2bar])
+
+        xbar_reverse = cg.independentFunctionList[0].xbar
+        ybar_reverse = cg.independentFunctionList[1].xbar
+
+        xbar_symbolic = 3 * x**2 * y**2 + y**2
+        ybar_symbolic = 2*x**3 * y + 2 * x * y
+
+        assert_almost_equal(xbar_reverse.data, xbar_symbolic.data)
+        assert_almost_equal(ybar_reverse.data, ybar_symbolic.data)
+
+
     def test_pullback_inv(self):
         """
         test pullback on
@@ -971,27 +1000,31 @@ class Test_CGgraph_on_UTPM(TestCase):
 
         assert_array_almost_equal(xbar_correct.data, fx.xbar.data)
 
-
     def test_broadcasting1(self):
-        D,P,N = 2,1,2
-        x = UTPM(numpy.random.rand(D,P,N,1))
-        A = UTPM(numpy.random.rand(D,P,N, N))
+        D, P, N = 2, 1, 2
+        x = UTPM(numpy.arange(D * P * N * 1, dtype=float).reshape((D, P, N, 1)))
+        A = UTPM(numpy.arange(D * P * N * N, dtype=float).reshape((D, P, N, N)))
         cg = CGraph()
         x = Function(x)
         A = Function(A)
         z = x + A
         cg.trace_off()
-        cg.independentFunctionList = [x,A]
+        cg.independentFunctionList = [x, A]
         cg.dependentFunctionList = [z]
 
-        zbar = UTPM(numpy.random.rand(*z.x.data.shape))
+        zbar = UTPM(3 + numpy.arange(D * P * N * N,
+                                 dtype=float).reshape((D, P, N, N)))
         cg.pullback([zbar])
 
-        print zbar.data[0,0]
-        print A.xbar.data[0,0]
-        print x.xbar.data[0,0]
-        assert_array_almost_equal(numpy.sum(z.x.data[1,0] * zbar.data[0,0]), numpy.sum(x.x.data[1,0] * x.xbar.data[0,0]) + numpy.sum(A.x.data[1,0] * A.xbar.data[0,0]))
+        # print 'zbar.data[0, 0]\n', zbar.data[0, 0]
+        # print 'A.xbar.data[0, 0]\n', A.xbar.data[0, 0]
+        # print 'x.xbar.data[0, 0]\n', x.xbar.data[0, 0]
 
+        # print cg
+        a = numpy.sum(z.x.data[1, 0] * zbar.data[0, 0])
+        b = numpy.sum(x.x.data[1, 0] * x.xbar.data[0, 0]) \
+          + numpy.sum(A.x.data[1, 0] * A.xbar.data[0, 0])
+        assert_array_almost_equal(a, b)
 
     def test_broadcasting2(self):
         D,P,N = 2,1,2
@@ -1011,6 +1044,28 @@ class Test_CGgraph_on_UTPM(TestCase):
         assert_array_almost_equal(numpy.sum(z.x.data[1,0] * zbar.data[0,0]), numpy.sum(x.x.data[1,0] * x.xbar.data[0,0]) + numpy.sum(A.x.data[1,0] * A.xbar.data[0,0]))
 
     def test_broadcasting3(self):
+        D, P, N = 2, 1, 2
+        x = UTPM(numpy.random.rand(D, P, N, 1))
+        A = UTPM(numpy.random.rand(D, P, N, N))
+        cg = CGraph()
+        x = Function(x)
+        A = Function(A)
+        #z = A - dot(x,x.T)/A + A*x /dot(A[:,:1], A[1:,:])
+        z = A * x
+        cg.trace_off()
+        cg.independentFunctionList = [x, A]
+        cg.dependentFunctionList = [z]
+
+        zbar = UTPM(numpy.random.rand(*z.x.data.shape))
+        cg.pullback([zbar])
+
+        a = numpy.sum(z.x.data[1, 0] * zbar.data[0, 0])
+        b = numpy.sum(x.x.data[1, 0] * x.xbar.data[0, 0]) \
+          + numpy.sum(A.x.data[1, 0] * A.xbar.data[0, 0])
+
+        assert_array_almost_equal(a, b)
+
+    def test_broadcasting4(self):
         D,P = 2,1
         x = UTPM(numpy.random.rand(D,P,2,3))
         y = UTPM(numpy.random.rand(D,P,3))
@@ -1025,11 +1080,32 @@ class Test_CGgraph_on_UTPM(TestCase):
         zbar = UTPM(numpy.random.rand(*z.x.data.shape))
         cg.pullback([zbar])
 
-        print zbar.data[0,0]
-        print y.xbar.data[0,0]
-        print x.xbar.data[0,0]
+        #print zbar.data[0,0]
+        #print y.xbar.data[0,0]
+        #print x.xbar.data[0,0]
         assert_array_almost_equal(numpy.sum(z.x.data[1,0] * zbar.data[0,0]), numpy.sum(x.x.data[1,0] * x.xbar.data[0,0]) + numpy.sum(y.x.data[1,0] * y.xbar.data[0,0]))
 
+    def test_broadcasting5(self):
+        D, P, N = 2, 1, 2
+        x = UTPM(numpy.random.rand(D, P, N, 1))
+        A = UTPM(numpy.random.rand(D, P, N, N))
+        cg = CGraph()
+        x = Function(x)
+        A = Function(A)
+        z = A - dot(x, x.T) / A + A * x / dot(A[:, :1], A[1:, :])
+        cg.trace_off()
+        cg.independentFunctionList = [x, A]
+        cg.dependentFunctionList = [z]
+
+        zbar = UTPM(numpy.random.rand(*z.x.data.shape))
+        cg.pullback([zbar])
+
+        a = numpy.sum(z.x.data[1, 0] * zbar.data[0, 0])
+        b = numpy.sum(x.x.data[1, 0] * x.xbar.data[0, 0]) \
+          + numpy.sum(A.x.data[1, 0] * A.xbar.data[0, 0])
+
+        assert_array_almost_equal(a, b)
+
     def test_eigh1_pullback(self):
         (D,P,N) = 2,1,2
         A = UTPM(numpy.zeros((D,P,N,N)))
@@ -1163,60 +1239,756 @@ class Test_CGgraph_on_UTPM(TestCase):
         assert_array_almost_equal(const1.data[0,:], const2.data[0,:])
 
 
-
-    def test_simple_ODOE_objective_function(self):
+    @decorators.skipif(mpmath is None)
+    def test_dpm_hyp1f1(self):
         """
-        compute PHI = trace( (J^T,J)^-1 )
+        compute y = dpm_hyp1f1(1., 2., x**2 + 3.)
         """
-        D,P,N,M = 2,1,3,2
-        MJ = UTPM(numpy.random.rand(D,P,N,M))
-       #  MJ = UTPM(numpy.array([[[[ 0.79636398,  0.22627532],
-       #   [ 0.89051158,  0.25639902],
-       #   [ 0.11028533,  0.95436784]]],
-
-
-       # [[[ 0.87357288,  0.12196627],
-       #   [ 0.77661888,  0.81793605],
-       #   [ 0.96229905,  0.7880918 ]]]]))
 
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.dpm_hyp1f1(1., 2., v1)
+            return y
 
+        # use CGraph
 
         cg = CGraph()
-        FJ = Function(MJ)
-        # print 'FJ=\n',FJ
-        FJT = Function.transpose(FJ)
-        # print 'FJT=\n',FJT
-        FM = Function(UTPM(numpy.zeros((D,P,M,M))))
-        # print 'FM=\n',FM
-        FM[...] += Function.dot(FJT, FJ)
-        # print 'FM=\n',FM
-        FC = Function.inv(FM)
-        # print 'FC=\n',FC
-        FPHI = Function.trace(FC)
-        cg.independentFunctionList = [FJ]
-        cg.dependentFunctionList = [FPHI]
+        x = Function(numpy.array([1.]))
+        y = f(x)
 
-        assert_array_equal(FPHI.shape, ())
-        PHIbar = UTPM(numpy.ones((D,P)))
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
 
-        # pullback using the tracer
-        cg.pullback([PHIbar])
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
 
-        # # compute pullback by hand
-        # Cbar = UTPM.pb_trace(PHIbar, FC.x, FPHI.x)
-        # assert_array_almost_equal(Cbar.data, FC.xbar.data)
+        # use UTPM
 
-        # Mbar = UTPM.pb_inv(Cbar, FM.x, FC.x)
-        # assert_array_almost_equal(Mbar.data, FM.xbar.data)
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
 
-        # tmpbar = UTPM.pb_dot(Mbar, FJT.x, FJ.x, FM.x)
-        # assert_array_almost_equal(tmpbar[1].data, FJ.xbar.data)
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
 
 
-        # verifying pullback by  ybar.T ydot == xbar.T xdot
-        const1 =  UTPM.dot(FPHI.xbar, UTPM.shift(FPHI.x,-1))
-        const2 = UTPM.trace(UTPM.dot(FJ.xbar.T, UTPM.shift(FJ.x,-1)))
-        assert_array_almost_equal(const1.data[0,:], const2.data[0,:])
+    def test_hyp1f1(self):
+        """
+        compute y = hyp1f1(1., 2., x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.hyp1f1(1., 2., v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_hyperu(self):
+        """
+        compute y = hyperu(1., 1.5, x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.hyperu(1., 1.5, v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    @decorators.skipif(mpmath is None)
+    def test_dpm_hyp2f0(self):
+        """
+        compute y = dpm_hyp2f0(0.5, 1.0, 0.1 * x**2 + 0.03)
+        """
+
+        def f(x):
+            # use smaller offset to ameliorate convergence issues
+            v1 = 0.1 * x**2 + 0.03
+            y = algopy.special.dpm_hyp2f0(0.5, 1.0, v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_hyp2f0(self):
+        """
+        compute y = hyp2f0(0.5, 1.0, 0.1 * x**2 + 0.03)
+        """
+
+        def f(x):
+            # use smaller offset to ameliorate convergence issues
+            v1 = 0.1 * x**2 + 0.03
+            y = algopy.special.hyp2f0(0.5, 1.0, v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_hyp0f1(self):
+        """
+        compute y = hyp0f1(2., x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.hyp0f1(2., v1)
+            return y
+
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_polygamma(self):
+        """
+        compute y = polygamma(2, x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.polygamma(2, v1)
+            return y
+
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_psi(self):
+        """
+        compute y = psi(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.psi(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_gammaln(self):
+        """
+        compute y = gammaln(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.gammaln(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_erf(self):
+        """
+        compute y = erf(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.erf(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_erfi(self):
+        """
+        compute y = erfi(x**2 - 3.)
+        """
+
+        def f(x):
+            v1 = x**2 - 3.
+            y = algopy.special.erfi(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_dawsn(self):
+        """
+        compute y = dawsn(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.dawsn(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_exp(self):
+        """
+        compute y = exp(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.exp(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_logit(self):
+        """
+        compute y = logit(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 0.1
+            y = algopy.special.logit(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_expit(self):
+        """
+        compute y = expit(x**2 + 3.)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.
+            y = algopy.special.expit(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_expm1(self):
+        """
+        compute y = expm1(x**2 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**2 + 3.0
+            y = algopy.expm1(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+
+    def test_log1p(self):
+        """
+        compute y = log1p(x**2 + 0.4)
+        """
+
+        def f(x):
+            v1 = x**2 + 0.4
+            y = algopy.log1p(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_absolute(self):
+        """
+        compute y = absolute(x**3 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**3 + 3.0
+            y = algopy.absolute(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_reciprocal(self):
+        """
+        compute y = reciprocal(x**3 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**3 + 3.0
+            y = algopy.reciprocal(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_square(self):
+        """
+        compute y = square(x**3 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**3 + 3.0
+            y = algopy.square(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_negative(self):
+        """
+        compute y = negative(x**3 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**3 + 3.0
+            y = algopy.negative(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_sign(self):
+        """
+        compute y = sign(x**3 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**3 + 3.0
+            y = algopy.sign(v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([0.2]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([0.2]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([0.2]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([0.2]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
+
+    def test_botched_clip(self):
+        """
+        compute y = botched_clip(1., 5., x**3 + 3.0)
+        """
+
+        def f(x):
+            v1 = x**3 + 3.
+            y = algopy.special.botched_clip(1., 5., v1)
+            return y
+
+        # use CGraph
+
+        cg = CGraph()
+        x = Function(numpy.array([1.]))
+        y = f(x)
+
+        cg.independentFunctionList = [x]
+        cg.dependentFunctionList = [y]
+
+        result1 = cg.jac_vec(numpy.array([2.]), numpy.array([1.]))
+        result2 = cg.jacobian(numpy.array([2.]))[0]
+
+        # use UTPM
+
+        x = UTPM.init_jacobian(numpy.array([2.]))
+        y = f(x)
+        result3 = UTPM.extract_jacobian(y)[0]
+
+        assert_array_almost_equal(result1, result2)
+        assert_array_almost_equal(result2, result3)
+        assert_array_almost_equal(result3, result1)
 
 
 
@@ -1266,70 +2038,6 @@ class Test_CGgraph_on_UTPM(TestCase):
 
         assert_array_almost_equal(const1.data[0,:], const2.data[0,:])
 
-    def test_complicated_ODOE_objective_function(self):
-
-        def Cfcn(F1p_list, out = None, work = None):
-            from numpy import sum, zeros
-            from algopy.globalfuncs import inv, dot, zeros
-
-            # check input arguments
-            Nex  = len(F1p_list)
-            Np   = F1p_list[0].shape[1]
-
-            # create temporary matrix M if not provided
-            # to store M = [[J_1^T J_1, J_2^T],[J_2, 0]]
-            if work == None:
-                work = zeros((Np,Np), dtype=F1p_list[0])
-            M = work
-
-            # Step 1:   compute M = J_1^T J_1
-            for nex in range(Nex):
-                M += dot(F1p_list[nex].T, F1p_list[nex])
-
-            # Step 2: invert M and prepare output
-
-            if out == None:
-                out = inv(M)
-            else:
-                out[...] = M
-            return out
-
-        D,P,N,M = 2,1,100,3
-        F1p_list = [UTPM(numpy.random.rand(D,P,N,M)),UTPM(numpy.random.rand(D,P,N,M))]
-        cg = CGraph()
-        FF1p_list = [Function(F1p) for F1p in F1p_list]
-        FC = Cfcn(FF1p_list)
-        FPHI = Function.trace(FC)
-
-        cg.independentFunctionList = FF1p_list
-        cg.dependentFunctionList = [FPHI]
-
-        assert_array_equal(FPHI.shape, ())
-        # cg.pushforward(F1p_list)
-        PHIbar = UTPM(numpy.zeros((D,P)))
-        PHIbar.data[0,:] = 1.
-
-        # pullback using the tracer
-        cg.pullback([PHIbar])
-
-        # verifying the computation
-        const1 =  UTPM.dot(PHIbar, UTPM.shift(FPHI.x,-1))
-        # print const1
-
-        Cbar = UTPM.pb_trace(PHIbar, FC.x, FPHI.x)
-        assert_array_almost_equal(Cbar.data, FC.xbar.data)
-        const2 =  UTPM.trace(UTPM.dot(Cbar.T, UTPM.shift(FC.x,-1)))
-        # print const2
-
-        const3 = UTPM(numpy.zeros((D,P)))
-
-        for nFF1p, FF1p in enumerate(FF1p_list):
-            const3 += UTPM.trace(UTPM.dot(FF1p.xbar.T, UTPM.shift(FF1p.x,-1)))
-
-        # print const3
-
-        assert_array_almost_equal(const1.data[0,:], const2.data[0,:])
-        assert_array_almost_equal(const2.data[0,:], const3.data[0,:])
 
     def test_simple_repeated_buffered_operation(self):
         """
@@ -1549,10 +2257,6 @@ class Test_UserFriendlyDrivers(TestCase):
 
 
 
-
-
-
-
     def test_jacobian_vec_hess(self):
         class Model:
             def eval_g(self, x):
@@ -1596,6 +2300,42 @@ class Test_UserFriendlyDrivers(TestCase):
         assert_array_almost_equal(m.eval_vec_hess_g_forward(w,x), m.eval_vec_hess_g_reverse(w,x))
 
 
+    def test_taylor_series_of_jacobian(self):
+
+        def eval_g(x):
+            out = algopy.zeros(2, dtype=x)
+            out[0] = algopy.sin(x[0]*x[1])
+            out[1] = algopy.exp(x[0]*algopy.cos(x[0]))
+            return out[...]
+
+        def eval_J(x):
+            out = algopy.zeros((2,2), dtype=x)
+            out[0, 0] = x[1] * algopy.cos(x[0] * x[1])
+            out[0, 1] = x[0] * algopy.cos(x[0] * x[1])
+            out[1, 0] = (algopy.cos(x[0]) - x[0] * algopy.sin(x[0])) * \
+                        algopy.exp( x[0] * algopy.cos(x[0]))
+            return out
+
+        D, P, N = 5, 3, 2
+
+        x = numpy.random.random((N))
+
+        cg = algopy.CGraph()
+        fx = algopy.Function(x)
+        fy = eval_g(fx)
+        cg.independentFunctionList = [fx]
+        cg.dependentFunctionList = [fy]
+
+        ax = numpy.zeros((D, P, N))
+        ax[...] = x.reshape((1, 1, N))
+        ax = algopy.UTPM(ax)
+
+        aJ = cg.jacobian(ax)
+
+        aJ2 = eval_J(ax)
+
+        assert_almost_equal(aJ.data, aJ2.data)
+
 
 class Test_CGraph_Plotting(TestCase):
     def test_simple(self):
